@@ -3,7 +3,8 @@
 use std::sync::Arc;
 
 use cov_mark::hit;
-use log::debug;
+use syntax::SmolStr;
+use tracing::debug;
 
 use chalk_ir::{cast::Cast, fold::shift::Shift, CanonicalVarKinds};
 use chalk_solve::rust_ir::{self, OpaqueTyDatumBound, WellKnownTrait};
@@ -11,14 +12,15 @@ use chalk_solve::rust_ir::{self, OpaqueTyDatumBound, WellKnownTrait};
 use base_db::CrateId;
 use hir_def::{
     lang_item::{lang_attr, LangItemTarget},
-    AssocContainerId, AssocItemId, GenericDefId, HasModule, Lookup, ModuleId, TypeAliasId,
+    AssocItemId, GenericDefId, HasModule, ItemContainerId, Lookup, ModuleId, TypeAliasId,
 };
 use hir_expand::name::name;
 
 use crate::{
     db::HirDatabase,
     display::HirDisplay,
-    from_assoc_type_id, from_chalk_trait_id, from_foreign_def_id, make_only_type_binders,
+    from_assoc_type_id, from_chalk_trait_id, from_foreign_def_id, make_binders,
+    make_single_type_binders,
     mapping::{from_chalk, ToChalk, TypeAliasAsValue},
     method_resolution::{TraitImpls, TyFingerprint, ALL_FLOAT_FPS, ALL_INT_FPS},
     to_assoc_type_id, to_chalk_trait_id,
@@ -60,7 +62,7 @@ impl<'a> chalk_solve::RustIrDatabase<Interner> for ChalkContext<'a> {
     }
     fn discriminant_type(&self, _ty: chalk_ir::Ty<Interner>) -> chalk_ir::Ty<Interner> {
         // FIXME: keep track of this
-        chalk_ir::TyKind::Scalar(chalk_ir::Scalar::Uint(chalk_ir::UintTy::U32)).intern(&Interner)
+        chalk_ir::TyKind::Scalar(chalk_ir::Scalar::Uint(chalk_ir::UintTy::U32)).intern(Interner)
     }
     fn impl_datum(&self, impl_id: ImplId) -> Arc<ImplDatum> {
         self.db.impl_datum(self.krate, impl_id)
@@ -82,14 +84,14 @@ impl<'a> chalk_solve::RustIrDatabase<Interner> for ChalkContext<'a> {
         debug!("impls_for_trait {:?}", trait_id);
         let trait_: hir_def::TraitId = from_chalk_trait_id(trait_id);
 
-        let ty: Ty = parameters[0].assert_ty_ref(&Interner).clone();
+        let ty: Ty = parameters[0].assert_ty_ref(Interner).clone();
 
         fn binder_kind(
             ty: &Ty,
             binders: &CanonicalVarKinds<Interner>,
         ) -> Option<chalk_ir::TyVariableKind> {
-            if let TyKind::BoundVar(bv) = ty.kind(&Interner) {
-                let binders = binders.as_slice(&Interner);
+            if let TyKind::BoundVar(bv) = ty.kind(Interner) {
+                let binders = binders.as_slice(Interner);
                 if bv.debruijn == DebruijnIndex::INNERMOST {
                     if let chalk_ir::VariableKind::Ty(tk) = binders[bv.index].kind {
                         return Some(tk);
@@ -172,8 +174,8 @@ impl<'a> chalk_solve::RustIrDatabase<Interner> for ChalkContext<'a> {
         // We don't do coherence checking (yet)
         unimplemented!()
     }
-    fn interner(&self) -> &Interner {
-        &Interner
+    fn interner(&self) -> Interner {
+        Interner
     }
     fn well_known_trait_id(
         &self,
@@ -205,18 +207,15 @@ impl<'a> chalk_solve::RustIrDatabase<Interner> for ChalkContext<'a> {
                 let (datas, binders) = (*datas).as_ref().into_value_and_skipped_binders();
                 let data = &datas.impl_traits[idx as usize];
                 let bound = OpaqueTyDatumBound {
-                    bounds: make_only_type_binders(
-                        1,
-                        data.bounds.skip_binders().iter().cloned().collect(),
-                    ),
-                    where_clauses: make_only_type_binders(0, vec![]),
+                    bounds: make_single_type_binders(data.bounds.skip_binders().to_vec()),
+                    where_clauses: chalk_ir::Binders::empty(Interner, vec![]),
                 };
                 chalk_ir::Binders::new(binders, bound)
             }
             crate::ImplTraitId::AsyncBlockTypeImplTrait(..) => {
                 if let Some((future_trait, future_output)) = self
                     .db
-                    .lang_item(self.krate, "future_trait".into())
+                    .lang_item(self.krate, SmolStr::new_inline("future_trait"))
                     .and_then(|item| item.as_trait())
                     .and_then(|trait_| {
                         let alias =
@@ -234,12 +233,12 @@ impl<'a> chalk_solve::RustIrDatabase<Interner> for ChalkContext<'a> {
                         trait_id: to_chalk_trait_id(future_trait),
                         // Self type as the first parameter.
                         substitution: Substitution::from1(
-                            &Interner,
+                            Interner,
                             TyKind::BoundVar(BoundVar {
                                 debruijn: DebruijnIndex::INNERMOST,
                                 index: 0,
                             })
-                            .intern(&Interner),
+                            .intern(Interner),
                         ),
                     });
                     let proj_bound = WhereClause::AliasEq(AliasEq {
@@ -247,35 +246,32 @@ impl<'a> chalk_solve::RustIrDatabase<Interner> for ChalkContext<'a> {
                             associated_ty_id: to_assoc_type_id(future_output),
                             // Self type as the first parameter.
                             substitution: Substitution::from1(
-                                &Interner,
+                                Interner,
                                 TyKind::BoundVar(BoundVar::new(DebruijnIndex::INNERMOST, 0))
-                                    .intern(&Interner),
+                                    .intern(Interner),
                             ),
                         }),
                         // The parameter of the opaque type.
                         ty: TyKind::BoundVar(BoundVar { debruijn: DebruijnIndex::ONE, index: 0 })
-                            .intern(&Interner),
+                            .intern(Interner),
                     });
                     let bound = OpaqueTyDatumBound {
-                        bounds: make_only_type_binders(
-                            1,
-                            vec![
-                                crate::wrap_empty_binders(impl_bound),
-                                crate::wrap_empty_binders(proj_bound),
-                            ],
-                        ),
-                        where_clauses: make_only_type_binders(0, vec![]),
+                        bounds: make_single_type_binders(vec![
+                            crate::wrap_empty_binders(impl_bound),
+                            crate::wrap_empty_binders(proj_bound),
+                        ]),
+                        where_clauses: chalk_ir::Binders::empty(Interner, vec![]),
                     };
                     // The opaque type has 1 parameter.
-                    make_only_type_binders(1, bound)
+                    make_single_type_binders(bound)
                 } else {
                     // If failed to find Symbol’s value as variable is void: Future::Output, return empty bounds as fallback.
                     let bound = OpaqueTyDatumBound {
-                        bounds: make_only_type_binders(0, vec![]),
-                        where_clauses: make_only_type_binders(0, vec![]),
+                        bounds: chalk_ir::Binders::empty(Interner, vec![]),
+                        where_clauses: chalk_ir::Binders::empty(Interner, vec![]),
                     };
                     // The opaque type has 1 parameter.
-                    make_only_type_binders(1, bound)
+                    make_single_type_binders(bound)
                 }
             }
         };
@@ -285,7 +281,7 @@ impl<'a> chalk_solve::RustIrDatabase<Interner> for ChalkContext<'a> {
 
     fn hidden_opaque_type(&self, _id: chalk_ir::OpaqueTyId<Interner>) -> chalk_ir::Ty<Interner> {
         // FIXME: actually provide the hidden type; it is relevant for auto traits
-        TyKind::Error.intern(&Interner)
+        TyKind::Error.intern(Interner)
     }
 
     fn is_object_safe(&self, _trait_id: chalk_ir::TraitId<Interner>) -> bool {
@@ -306,13 +302,13 @@ impl<'a> chalk_solve::RustIrDatabase<Interner> for ChalkContext<'a> {
         _closure_id: chalk_ir::ClosureId<Interner>,
         substs: &chalk_ir::Substitution<Interner>,
     ) -> chalk_ir::Binders<rust_ir::FnDefInputsAndOutputDatum<Interner>> {
-        let sig_ty = substs.at(&Interner, 0).assert_ty_ref(&Interner).clone();
+        let sig_ty = substs.at(Interner, 0).assert_ty_ref(Interner).clone();
         let sig = &sig_ty.callable_sig(self.db).expect("first closure param should be fn ptr");
         let io = rust_ir::FnDefInputsAndOutputDatum {
-            argument_types: sig.params().iter().cloned().collect(),
+            argument_types: sig.params().to_vec(),
             return_type: sig.ret().clone(),
         };
-        make_only_type_binders(0, io.shifted_in(&Interner))
+        chalk_ir::Binders::empty(Interner, io.shifted_in(Interner))
     }
     fn closure_upvars(
         &self,
@@ -320,14 +316,14 @@ impl<'a> chalk_solve::RustIrDatabase<Interner> for ChalkContext<'a> {
         _substs: &chalk_ir::Substitution<Interner>,
     ) -> chalk_ir::Binders<chalk_ir::Ty<Interner>> {
         let ty = TyBuilder::unit();
-        make_only_type_binders(0, ty)
+        chalk_ir::Binders::empty(Interner, ty)
     }
     fn closure_fn_substitution(
         &self,
         _closure_id: chalk_ir::ClosureId<Interner>,
         _substs: &chalk_ir::Substitution<Interner>,
     ) -> chalk_ir::Substitution<Interner> {
-        Substitution::empty(&Interner)
+        Substitution::empty(Interner)
     }
 
     fn trait_name(&self, trait_id: chalk_ir::TraitId<Interner>) -> String {
@@ -341,6 +337,10 @@ impl<'a> chalk_solve::RustIrDatabase<Interner> for ChalkContext<'a> {
             hir_def::AdtId::UnionId(id) => self.db.union_data(id).name.to_string(),
         }
     }
+    fn adt_size_align(&self, _id: chalk_ir::AdtId<Interner>) -> Arc<rust_ir::AdtSizeAlign> {
+        // FIXME
+        Arc::new(rust_ir::AdtSizeAlign::from_one_zst(false))
+    }
     fn assoc_type_name(&self, assoc_ty_id: chalk_ir::AssocTypeId<Interner>) -> String {
         let id = self.db.associated_ty_data(assoc_ty_id).name;
         self.db.type_alias_data(id).name.to_string()
@@ -399,7 +399,7 @@ pub(crate) fn associated_ty_data_query(
     debug!("associated_ty_data {:?}", id);
     let type_alias: TypeAliasId = from_assoc_type_id(id);
     let trait_ = match type_alias.lookup(db.upcast()).container {
-        AssocContainerId::TraitId(t) => t,
+        ItemContainerId::TraitId(t) => t,
         _ => panic!("associated type not in trait"),
     };
 
@@ -409,16 +409,31 @@ pub(crate) fn associated_ty_data_query(
     // let bound_vars = generic_params.bound_vars_subst(DebruijnIndex::INNERMOST);
     let resolver = hir_def::resolver::HasResolver::resolver(type_alias, db.upcast());
     let ctx = crate::TyLoweringContext::new(db, &resolver)
-        .with_type_param_mode(crate::lower::TypeParamLoweringMode::Variable);
+        .with_type_param_mode(crate::lower::ParamLoweringMode::Variable);
     let self_ty =
-        TyKind::BoundVar(BoundVar::new(crate::DebruijnIndex::INNERMOST, 0)).intern(&Interner);
-    let bounds = type_alias_data
+        TyKind::BoundVar(BoundVar::new(crate::DebruijnIndex::INNERMOST, 0)).intern(Interner);
+    let mut bounds: Vec<_> = type_alias_data
         .bounds
         .iter()
         .flat_map(|bound| ctx.lower_type_bound(bound, self_ty.clone(), false))
         .filter_map(|pred| generic_predicate_to_inline_bound(db, &pred, &self_ty))
         .collect();
 
+    if !ctx.unsized_types.borrow().contains(&self_ty) {
+        let sized_trait = resolver
+            .krate()
+            .and_then(|krate| db.lang_item(krate, SmolStr::new_inline("sized")))
+            .and_then(|lang_item| lang_item.as_trait().map(to_chalk_trait_id));
+        let sized_bound = sized_trait.into_iter().map(|sized_trait| {
+            let trait_bound =
+                rust_ir::TraitBound { trait_id: sized_trait, args_no_self: Default::default() };
+            let inline_bound = rust_ir::InlineBound::TraitBound(trait_bound);
+            chalk_ir::Binders::empty(Interner, inline_bound)
+        });
+        bounds.extend(sized_bound);
+        bounds.shrink_to_fit();
+    }
+
     // FIXME: Re-enable where clauses on associated types when an upstream chalk bug is fixed.
     //        (rust-analyzer#9052)
     // let where_clauses = convert_where_clauses(db, type_alias.into(), &bound_vars);
@@ -427,7 +442,7 @@ pub(crate) fn associated_ty_data_query(
         trait_id: to_chalk_trait_id(trait_),
         id,
         name: type_alias,
-        binders: make_only_type_binders(generic_params.len(), bound_data),
+        binders: make_binders(db, &generic_params, bound_data),
     };
     Arc::new(datum)
 }
@@ -442,7 +457,7 @@ pub(crate) fn trait_datum_query(
     let trait_data = db.trait_data(trait_);
     debug!("trait {:?} = {:?}", trait_id, trait_data.name);
     let generic_params = generics(db.upcast(), trait_.into());
-    let bound_vars = generic_params.bound_vars_subst(DebruijnIndex::INNERMOST);
+    let bound_vars = generic_params.bound_vars_subst(db, DebruijnIndex::INNERMOST);
     let flags = rust_ir::TraitFlags {
         auto: trait_data.is_auto,
         upstream: trait_.lookup(db.upcast()).container.krate() != krate,
@@ -459,7 +474,7 @@ pub(crate) fn trait_datum_query(
         lang_attr(db.upcast(), trait_).and_then(|name| well_known_trait_from_lang_attr(&name));
     let trait_datum = TraitDatum {
         id: trait_id,
-        binders: make_only_type_binders(bound_vars.len(&Interner), trait_datum_bound),
+        binders: make_binders(db, &generic_params, trait_datum_bound),
         flags,
         associated_ty_ids,
         well_known,
@@ -469,33 +484,38 @@ pub(crate) fn trait_datum_query(
 
 fn well_known_trait_from_lang_attr(name: &str) -> Option<WellKnownTrait> {
     Some(match name {
-        "sized" => WellKnownTrait::Sized,
-        "copy" => WellKnownTrait::Copy,
         "clone" => WellKnownTrait::Clone,
-        "drop" => WellKnownTrait::Drop,
-        "fn_once" => WellKnownTrait::FnOnce,
-        "fn_mut" => WellKnownTrait::FnMut,
-        "fn" => WellKnownTrait::Fn,
-        "unsize" => WellKnownTrait::Unsize,
         "coerce_unsized" => WellKnownTrait::CoerceUnsized,
+        "copy" => WellKnownTrait::Copy,
         "discriminant_kind" => WellKnownTrait::DiscriminantKind,
+        "dispatch_from_dyn" => WellKnownTrait::DispatchFromDyn,
+        "drop" => WellKnownTrait::Drop,
+        "fn" => WellKnownTrait::Fn,
+        "fn_mut" => WellKnownTrait::FnMut,
+        "fn_once" => WellKnownTrait::FnOnce,
+        "generator" => WellKnownTrait::Generator,
+        "sized" => WellKnownTrait::Sized,
+        "unpin" => WellKnownTrait::Unpin,
+        "unsize" => WellKnownTrait::Unsize,
         _ => return None,
     })
 }
 
 fn lang_attr_from_well_known_trait(attr: WellKnownTrait) -> &'static str {
     match attr {
-        WellKnownTrait::Sized => "sized",
-        WellKnownTrait::Copy => "copy",
         WellKnownTrait::Clone => "clone",
-        WellKnownTrait::Drop => "drop",
-        WellKnownTrait::FnOnce => "fn_once",
-        WellKnownTrait::FnMut => "fn_mut",
-        WellKnownTrait::Fn => "fn",
-        WellKnownTrait::Unsize => "unsize",
-        WellKnownTrait::Unpin => "unpin",
         WellKnownTrait::CoerceUnsized => "coerce_unsized",
+        WellKnownTrait::Copy => "copy",
         WellKnownTrait::DiscriminantKind => "discriminant_kind",
+        WellKnownTrait::DispatchFromDyn => "dispatch_from_dyn",
+        WellKnownTrait::Drop => "drop",
+        WellKnownTrait::Fn => "fn",
+        WellKnownTrait::FnMut => "fn_mut",
+        WellKnownTrait::FnOnce => "fn_once",
+        WellKnownTrait::Generator => "generator",
+        WellKnownTrait::Sized => "sized",
+        WellKnownTrait::Unpin => "unpin",
+        WellKnownTrait::Unsize => "unsize",
     }
 }
 
@@ -506,11 +526,11 @@ pub(crate) fn struct_datum_query(
 ) -> Arc<StructDatum> {
     debug!("struct_datum {:?}", struct_id);
     let chalk_ir::AdtId(adt_id) = struct_id;
-    let num_params = generics(db.upcast(), adt_id.into()).len();
+    let generic_params = generics(db.upcast(), adt_id.into());
     let upstream = adt_id.module(db.upcast()).krate() != krate;
     let where_clauses = {
         let generic_params = generics(db.upcast(), adt_id.into());
-        let bound_vars = generic_params.bound_vars_subst(DebruijnIndex::INNERMOST);
+        let bound_vars = generic_params.bound_vars_subst(db, DebruijnIndex::INNERMOST);
         convert_where_clauses(db, adt_id.into(), &bound_vars)
     };
     let flags = rust_ir::AdtFlags {
@@ -528,7 +548,7 @@ pub(crate) fn struct_datum_query(
         // FIXME set ADT kind
         kind: rust_ir::AdtKind::Struct,
         id: struct_id,
-        binders: make_only_type_binders(num_params, struct_datum_bound),
+        binders: make_binders(db, &generic_params, struct_datum_bound),
         flags,
     };
     Arc::new(struct_datum)
@@ -560,7 +580,7 @@ fn impl_def_datum(
     let impl_data = db.impl_data(impl_id);
 
     let generic_params = generics(db.upcast(), impl_id.into());
-    let bound_vars = generic_params.bound_vars_subst(DebruijnIndex::INNERMOST);
+    let bound_vars = generic_params.bound_vars_subst(db, DebruijnIndex::INNERMOST);
     let trait_ = trait_ref.hir_trait_id();
     let impl_type = if impl_id.lookup(db.upcast()).container.krate() == krate {
         rust_ir::ImplType::Local
@@ -597,7 +617,7 @@ fn impl_def_datum(
         .collect();
     debug!("impl_datum: {:?}", impl_datum_bound);
     let impl_datum = ImplDatum {
-        binders: make_only_type_binders(bound_vars.len(&Interner), impl_datum_bound),
+        binders: make_binders(db, &generic_params, impl_datum_bound),
         impl_type,
         polarity,
         associated_ty_value_ids,
@@ -621,7 +641,7 @@ fn type_alias_associated_ty_value(
 ) -> Arc<AssociatedTyValue> {
     let type_alias_data = db.type_alias_data(type_alias);
     let impl_id = match type_alias.lookup(db.upcast()).container {
-        AssocContainerId::ImplId(it) => it,
+        ItemContainerId::ImplId(it) => it,
         _ => panic!("assoc ty value should be in impl"),
     };
 
@@ -653,17 +673,17 @@ pub(crate) fn fn_def_datum_query(
     let callable_def: CallableDefId = from_chalk(db, fn_def_id);
     let generic_params = generics(db.upcast(), callable_def.into());
     let (sig, binders) = db.callable_item_signature(callable_def).into_value_and_skipped_binders();
-    let bound_vars = generic_params.bound_vars_subst(DebruijnIndex::INNERMOST);
+    let bound_vars = generic_params.bound_vars_subst(db, DebruijnIndex::INNERMOST);
     let where_clauses = convert_where_clauses(db, callable_def.into(), &bound_vars);
     let bound = rust_ir::FnDefDatumBound {
         // Note: Chalk doesn't actually use this information yet as far as I am aware, but we provide it anyway
-        inputs_and_output: make_only_type_binders(
-            0,
+        inputs_and_output: chalk_ir::Binders::empty(
+            Interner,
             rust_ir::FnDefInputsAndOutputDatum {
-                argument_types: sig.params().iter().cloned().collect(),
+                argument_types: sig.params().to_vec(),
                 return_type: sig.ret().clone(),
             }
-            .shifted_in(&Interner),
+            .shifted_in(Interner),
         ),
         where_clauses,
     };
@@ -679,7 +699,7 @@ pub(crate) fn fn_def_variance_query(db: &dyn HirDatabase, fn_def_id: FnDefId) ->
     let callable_def: CallableDefId = from_chalk(db, fn_def_id);
     let generic_params = generics(db.upcast(), callable_def.into());
     Variances::from_iter(
-        &Interner,
+        Interner,
         std::iter::repeat(chalk_ir::Variance::Invariant).take(generic_params.len()),
     )
 }
@@ -690,7 +710,7 @@ pub(crate) fn adt_variance_query(
 ) -> Variances {
     let generic_params = generics(db.upcast(), adt_id.into());
     Variances::from_iter(
-        &Interner,
+        Interner,
         std::iter::repeat(chalk_ir::Variance::Invariant).take(generic_params.len()),
     )
 }
@@ -703,7 +723,7 @@ pub(super) fn convert_where_clauses(
     let generic_predicates = db.generic_predicates(def);
     let mut result = Vec::with_capacity(generic_predicates.len());
     for pred in generic_predicates.iter() {
-        result.push(pred.clone().substitute(&Interner, substs));
+        result.push(pred.clone().substitute(Interner, substs));
     }
     result
 }
@@ -715,30 +735,30 @@ pub(super) fn generic_predicate_to_inline_bound(
 ) -> Option<chalk_ir::Binders<rust_ir::InlineBound<Interner>>> {
     // An InlineBound is like a GenericPredicate, except the self type is left out.
     // We don't have a special type for this, but Chalk does.
-    let self_ty_shifted_in = self_ty.clone().shifted_in_from(&Interner, DebruijnIndex::ONE);
+    let self_ty_shifted_in = self_ty.clone().shifted_in_from(Interner, DebruijnIndex::ONE);
     let (pred, binders) = pred.as_ref().into_value_and_skipped_binders();
     match pred {
         WhereClause::Implemented(trait_ref) => {
-            if trait_ref.self_type_parameter(&Interner) != self_ty_shifted_in {
+            if trait_ref.self_type_parameter(Interner) != self_ty_shifted_in {
                 // we can only convert predicates back to type bounds if they
                 // have the expected self type
                 return None;
             }
-            let args_no_self = trait_ref.substitution.as_slice(&Interner)[1..]
+            let args_no_self = trait_ref.substitution.as_slice(Interner)[1..]
                 .iter()
-                .map(|ty| ty.clone().cast(&Interner))
+                .map(|ty| ty.clone().cast(Interner))
                 .collect();
             let trait_bound = rust_ir::TraitBound { trait_id: trait_ref.trait_id, args_no_self };
             Some(chalk_ir::Binders::new(binders, rust_ir::InlineBound::TraitBound(trait_bound)))
         }
         WhereClause::AliasEq(AliasEq { alias: AliasTy::Projection(projection_ty), ty }) => {
-            if projection_ty.self_type_parameter(&Interner) != self_ty_shifted_in {
+            if projection_ty.self_type_parameter(Interner) != self_ty_shifted_in {
                 return None;
             }
             let trait_ = projection_ty.trait_(db);
-            let args_no_self = projection_ty.substitution.as_slice(&Interner)[1..]
+            let args_no_self = projection_ty.substitution.as_slice(Interner)[1..]
                 .iter()
-                .map(|ty| ty.clone().cast(&Interner))
+                .map(|ty| ty.clone().cast(Interner))
                 .collect();
             let alias_eq_bound = rust_ir::AliasEqBound {
                 value: ty.clone(),
