@@ -1364,10 +1364,7 @@ fn fn_sig_for_fn(db: &dyn HirDatabase, def: FunctionId) -> PolyFnSig {
         .with_type_param_mode(ParamLoweringMode::Variable);
     let ret = ctx_ret.lower_ty(&data.ret_type);
     let generics = generics(db.upcast(), def.into());
-    let mut sig = CallableSig::from_params_and_return(params, ret, data.is_varargs());
-    if !data.legacy_const_generics_indices.is_empty() {
-        sig.set_legacy_const_generics_indices(&data.legacy_const_generics_indices);
-    }
+    let sig = CallableSig::from_params_and_return(params, ret, data.is_varargs());
     make_binders(db, &generics, sig)
 }
 
